import os
import ray
import time
from functools import partial
from collections import defaultdict, deque
from torch.utils.tensorboard import SummaryWriter

class Learner:
    """
    Task learner class.

    Class to manage the training process. It pushes forward training calling
    the update workers and tracks progress.

    Parameters
    ----------
    update_workers : Worker or WorkerSet
        The set of workers handling actor_critic updates.
    target_steps : int
        Number of environment steps to reach to complete training.
    log_dir : str
        Target directory for Tensorboard logs.

    Attributes
    ----------
    update_workers : Worker or WorkerSet
        The set of workers handling actor_critic updates.
    target_steps : int
        Number of environment steps to reach to complete training.
    log_dir : str
        Target directory for Tensorboard logs.
    num_samples_collected : int
        Number of train environment transition samples so far in the
        training process.
    metrics : collections.defaultdict
        Collection of deque's, tracking recent training metrics (as moving
        averages).
    start : float
        Training start time.
    """

    def __init__(self, update_workers, target_steps, log_dir=None):

        # Input attributes
        self.log_dir = log_dir
        self.update_workers = update_workers
        self.target_steps = target_steps

        # Counters and metrics
        self.num_samples_collected = 0
        self.metrics = defaultdict(partial(deque, maxlen=100))

        # Define summary writer
        if log_dir:
            tb_log_dir = "{}/tensorboard_logs".format(log_dir)
            os.makedirs(tb_log_dir, exist_ok=True)
            self.writer = SummaryWriter(log_dir=tb_log_dir)
        else: self.writer = None

        # Record starting time
        self.start = time.time()

    def step(self):
        """Takes a logical synchronous optimization step."""

        # Update step
        info = self.update_workers.step()
        info["scheme/metrics/fps"] = int(self.num_samples_collected / (time.time() - self.start))

        # Update counters
        self.num_samples_collected += info.pop("collected_samples")

        # Update and log metrics
        for k, v in info.items():
            self.metrics[k].append(v)
            if self.writer: self.writer.add_scalar(k, v, self.num_samples_collected)

    def done(self):
        """
        Return True if training has finished (target_steps reached).

        Returns
        -------
        flag : bool
            True if training has reached the target number of steps.
        """
        flag = self.num_samples_collected >= self.target_steps
        if flag: self.update_workers.stop()
        return flag

    def get_metrics(self):
        """Returns metrics averaged across number of updates."""
        return {k: sum(v) / len(v) for k, v in self.metrics.items()}

    def print_info(self):
        """Print relevant information about the training process"""
        s = "Update {}, num samples collected {}, FPS {}".format(
            self.update_workers.num_updates, self.num_samples_collected,
            int(self.num_samples_collected / (time.time() - self.start)))
        s += "\n"
        for k, v in self.get_metrics().items():
            if k.split("/")[0] == "algo":
                s += "{} {}, ".format(k.split("/")[-1], v)
        if "test" in self.get_metrics().keys():
            s += "test_perf {}, ".format(self.get_metrics()["test"][-1])
        print(s[:-2], flush=True)

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm used for training,
        change its value to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        self.update_workers.update_algo_parameter(parameter_name, new_parameter_value)

    def save_model(self):
        """
        Save currently learned actor_critic version.

        Returns
        -------
        save_name : str
            Path to saved file.
        """
        fname = os.path.join(self.log_dir, "actor_critic.state_dict")
        save_name = self.update_workers.save_model(fname)
        return save_name



import time
import torch
import numpy as np
from ray.services import get_node_ip_address

from ..base.worker import Worker as W
from .utils import check_message, find_free_port

class CGUWorker(W):
    """
    Worker class handling remote data collection, gradient computation and
    policy updates.

    This class wraps an actor_critic instance, a storage class instance and a
    train and a test vector of environments. It collects data, computes gradients,
    updates the networks and evaluates network versions following a logic
    defined in function self.step(), which will be called from the Learner.

    Parameters
    ----------
    index_worker : int
        Worker index.
    create_algo_instance : func
        A function that creates an algorithm class.
    create_storage_instance : func
        A function that create a rollouts storage.
    create_train_envs_instance : func
        A function to create train environments.
    create_actor_critic_instance : func
        A function that creates a policy.
    create_test_envs_instance : func
        A function to create test environments.
    initial_weights : ray object ID
        Initial model weights.

    Attributes
    ----------
    index_worker : int
        Index assigned to this worker.
    actor_critic : nn.Module
        An actor_critic class instance.
    algo : an algorithm class
        An algorithm class instance.
    envs_train : VecEnv
        A VecEnv class instance with the train environments.
    envs_test : VecEnv
        A VecEnv class instance with the test environments.
    storage : a rollout storage class
        A Storage class instance.
    iter : int
        Times actor critic has been updated.
    update_every : int
        Number of data samples to collect between network update stages.
    obs : torch.tensor
        Latest train environment observation.
    rhs : torch.tensor
        Latest policy recurrent hidden state.
    done : torch.tensor
        Latest train environment done flag.
    """

    def __init__(self,
                 index_worker,
                 create_algo_instance,
                 create_storage_instance,
                 create_train_envs_instance,
                 create_actor_critic_instance,
                 create_test_envs_instance=lambda x, y, c: None,
                 initial_weights=None):

        super(CGUWorker, self).__init__(index_worker)

        # Using Ray, worker should only see one GPU or None
        device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

        # Create Actor Critic instance
        self.actor_critic = create_actor_critic_instance(device)
        self.actor_critic.to(device)

        # Create Algorithm instance
        self.algo = create_algo_instance(self.actor_critic, device)

        # Define counters
        self.iter = 0

        if initial_weights: # if remote worker

            # Create train environments, define initial train states
            self.envs_train = create_train_envs_instance(device, index_worker)
            self.obs, self.rhs, self.done = self.actor_critic.policy_initial_states(self.envs_train.reset())

            # Create test environments (if creation function available)
            self.envs_test = create_test_envs_instance(device, index_worker, mode="test")

            # Create Storage instance and set world initial state
            self.storage = create_storage_instance(device)
            self.update_every = self.algo.update_every or self.storage.max_size

            # Print worker information
            self.print_worker_info()
            self.set_weights(initial_weights)

            # Collect initial samples
            print("Collecting initial samples...")
            self.collect_data(self.algo.start_steps, self.algo.start_steps)

    def collect_data(self, num_steps=None, min_fraction=1.0):
        """
        Collect data from interactions with the environments.

        Parameters
        ----------
        num_steps : int
            Target number of train environment steps to take.
        min_steps : int
            Minimum number of train environment steps to stop at if other
            workers have finished their collection task.
        """

        if self.iter % (self.algo.num_epochs * self.algo.num_mini_batch) != 0:
            return

        if num_steps is None:
            num_steps = int(self.update_every)

        min_steps = int(num_steps * min_fraction)

        t = time.time()
        for step in range(num_steps):

            # Predict next action, next rnn hidden state and algo-specific outputs
            act, clip_act, rhs, algo_data = self.algo.acting_step(self.obs, self.rhs, self.done)

            # Interact with envs_vector with predicted action (clipped within action space)
            obs2, reward, done, infos = self.envs_train.step(clip_act)

            # Prepare transition dict
            transition = {"obs": self.obs, "rhs": rhs, "act": act, "rew": reward, "obs2": obs2, "done": done}
            transition.update(algo_data)

            # Store transition in buffer
            self.storage.insert(transition)

            # Update current world state
            self.obs, self.rhs, self.done = obs2, rhs, done

            # Record model version used to collect data
            self.storage.ac_version = self.iter

            if check_message("sample") == b"stop" and step >= min_steps:
                self.num_steps_reached = step
                break

        # Record and return metrics
        self.collect_time = time.time() - t
        self.collect_samples = num_steps * self.envs_train.num_envs

    def compute_gradients(self, batch):
        """
        Calculate actor critic gradients and average them with grads from
        other workers.

        Parameters
        ----------
        batch : dict
            data batch containing all required tensors to compute algo loss.

        Returns
        -------
        info : dict
            Summary dict of relevant gradient-related information.
        """

        t = time.time()
        grads, info = self.algo.compute_gradients(batch, grads_to_cpu=False)
        compute_grads_t = time.time() - t

        ###### ALLREDUCE ######################################################

        t = time.time()
        if torch.cuda.is_available():
            for g in grads:
                torch.distributed.all_reduce(g, op=torch.distributed.ReduceOp.SUM)
        else:
            torch.distributed.all_reduce_coalesced(grads, op=torch.distributed.ReduceOp.SUM)

        for p in self.actor_critic.parameters():
            if p.grad is not None:
                p.grad /= self.distributed_world_size
        avg_grads_t = time.time() - t

        #######################################################################

        info.update({"scheme/seconds_to/compute_grads": compute_grads_t})
        info.update({"scheme/seconds_to/avg_grads_t": avg_grads_t})

        return info

    def update_networks(self):
        """Update Actor Critic model"""
        self.algo.apply_gradients()

    def evaluate(self):
        """
        Test current actor_critic version in self.envs_test.

        Returns
        -------
        mean_test_perf : float
            Average accumulated reward over all tested episodes.
        """

        completed_episodes = []
        obs = self.envs_test.reset()
        rewards = np.zeros(obs.shape[0])
        obs, rhs, done = self.actor_critic.policy_initial_states(obs)

        while len(completed_episodes) < self.algo.num_test_episodes:
            # Predict next action and rnn hidden state
            act, clip_act, rhs, _ = self.algo.acting_step(obs, rhs, done, deterministic=True)

            # Interact with env with predicted action (clipped within action space)
            obs2, reward, done, _ = self.envs_test.step(clip_act)

            # Keep track of episode rewards and completed episodes
            rewards += reward.cpu().squeeze(-1).numpy()
            completed_episodes.extend(
                rewards[done.cpu().squeeze(-1).numpy() == 1.0].tolist())
            rewards[done.cpu().squeeze(-1).numpy() == 1.0] = 0.0

            obs = obs2

        return np.mean(completed_episodes)

    def step(self):
        """
        Perform logical learning step. Training proceeds interleaving collection
        of data samples with gradient computations and policy updates.

        Returns
        -------
        info : dict
            Summary dict with relevant step information.

        """

        # Collect data and prepare data batches
        if self.iter % (self.algo.num_epochs * self.algo.num_mini_batch) == 0:
            self.storage.before_update(self.actor_critic, self.algo)
            self.batches = self.storage.generate_batches(
                self.algo.num_mini_batch, self.algo.mini_batch_size,
                self.algo.num_epochs, self.actor_critic.is_recurrent)
            collect_time, collected_samples = self.collect_time, self.collect_samples
        else:
            collect_time, collected_samples = 0.0, 0

        # Get next batch
        batch = self.batches.__next__()

        # Compute gradients
        info = self.compute_gradients(batch)

        # Update model
        self.update_networks()

        # Add extra information to info dict
        info.update({"ac_update_num": self.iter})
        info.update({"collected_samples": collected_samples})
        info.update({"scheme/seconds_to/collect": collect_time})
        info.update({"scheme/metrics/collection_gradient_delay": self.iter - self.storage.ac_version})
        if self.iter == 0: info["collected_samples"] += self.algo.start_steps * self.envs_train.num_envs

        # Update counter
        self.iter += 1

        # Evaluate current network
        if self.iter % self.algo.test_every == 0:
            if self.envs_test and self.algo.num_test_episodes > 0:
                test_perf = self.evaluate()
                info.update({"scheme/metrics/test_performance": test_perf})

        return info

    def setup_torch_data_parallel(self, url, world_rank, world_size, backend):
        """Join a torch process group for distributed SGD."""
        torch.distributed.init_process_group(
            backend=backend,
            init_method=url,
            rank=world_rank,
            world_size=world_size)
        self.distributed_world_size = world_size

    def get_node_ip(self):
        """Returns the IP address of the current node."""
        return get_node_ip_address()

    def find_free_port(self):
        """Returns a free port on the current node."""
        return find_free_port()

    def set_weights(self, weights):
        """
        Update the worker actor_critic version with provided weights.

        weights: dict of tensors
            Dict containing actor_critic weights to be set.
        """
        self.algo.set_weights(weights["weights"])

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """
        self.algo.update_algo_parameter(parameter_name, new_parameter_value)


import os
import ray
import time
import torch
from shutil import copy2
from functools import partial
from collections import defaultdict, deque
from .utils import compute_vtrace
from ..utils import TaskPool, ray_get_and_free


class GUWorker:
    """
    Update worker. Handles actor_critic updates.

    This class coordinates sequential central actor_critic optimization, using
    rollouts collected by distributed workers to compute gradients and update
    the models.

    Parameters
    ----------
    col_workers : WorkerSet
        Set of workers collecting and sending rollouts to the UWorker.
    device : torch.device
        CPU or specific GPU to use for computation.
    use_vtrace : bool
        Whether or not to use off-policy correction method called V-trace
        (https://arxiv.org/abs/1802.01561) to account for actor_critic version
        differences between collection and gradient computation tasks.
    broadcast_interval : int
        After how many central updates, model weights should be broadcasted to
        remote collection workers.
    max_collect_requests_pending : int
        maximum number of collection tasks simultaneously scheduled to each
        collection worker.

    Attributes
    ----------
    ps : Worker
        Local worker that acts as a parameter server.
    latest_weights : ray object ID
    col_workers : list of Workers
        Set of workers collecting and sending rollouts to the UWorker.
    num_updates : int
        Number of times the actor_critic model has been updated.
    num_workers : int
        number of remote workers computing gradients.
    use_vtrace: bool
        Whether or not to use off-policy correction method called V-trace
        (https://arxiv.org/abs/1802.01561) to account for actor_critic version
        differences between collection and gradient computation tasks.
    broadcast_interval : int
        After how many central updates, model weights should be broadcasted to
        remote collection workers.
    """
    def __init__(self,
                 col_workers,
                 device="cpu",
                 use_vtrace=False,
                 broadcast_interval=1,
                 max_collect_requests_pending=2):

        self.ps = col_workers.local_worker()
        self.ps.actor_critic.to(device)
        self.latest_weights = ray.put({"update": 0, "weights": self.ps.get_weights()})

        self.col_workers = col_workers.remote_workers()
        self.num_workers = len(self.col_workers)

        self.use_vtrace = use_vtrace
        self.broadcast_interval = broadcast_interval

        # Check remote workers exist
        if len(self.col_workers) == 0:
            raise ValueError("""At least 1 data collection worker required""")

        # Counters and metrics
        self.num_updates = 0
        self.num_sent_since_broadcast = 0
        self.metrics = defaultdict(partial(deque, maxlen=100))

        # Start collecting data
        self.collector_tasks = TaskPool()
        for ev in self.col_workers:
            for _ in range(max_collect_requests_pending):
                ev.set_weights.remote(self.latest_weights)
                self.collector_tasks.add(ev, ev.collect_data.remote())

    def compute_gradients(self, batch):
        """
        Calculate actor critic gradients.

        Parameters
        ----------
        batch : dict
            data batch containing all required tensors to compute algo loss.

        Returns
        -------
        info : dict
            Summary dict of relevant gradient-related information.
        """
        t = time.time()
        _, info = self.ps.algo.compute_gradients(batch)
        info.update({"scheme/seconds_to/compute_grads_t": time.time() - t})
        return info

    def update_networks(self):
        """Update Actor Critic model"""
        self.ps.algo.apply_gradients()

    def should_broadcast(self):
        """Returns whether broadcast() should be called to update weights"""
        return self.num_sent_since_broadcast >= self.broadcast_interval

    def broadcast_new_weights(self):
        """Broadcast a new set of weights from the local worker"""
        self.latest_weights = ray.put({
            "update": self.num_updates,
            "weights": self.ps.get_weights()})
        self.num_sent_since_broadcast = 0

    def step(self):
        """
        Takes a logical optimization step.

        Returns
        -------
        info : dict
            Summary dict of relevant step information.
        """

        n = self.ps.algo.num_epochs * self.ps.algo.num_mini_batch
        if self.num_updates % n == 0:

            # Wait to remote workers to complete data collection tasks
            for e, rollouts in self.collector_tasks.completed(blocking_wait=True,  max_yield=1):

                # Move new collected rollouts to parameter server input queue
                new_rollouts = ray_get_and_free(rollouts)

                # Update counter and broadcast weights to worker if necessary
                self.num_sent_since_broadcast += 1
                if self.should_broadcast():
                    self.broadcast_new_weights()

                # Request more data from worker
                e.set_weights.remote(self.latest_weights)
                self.collector_tasks.add(e, e.collect_data.remote())

            self.ps.storage.add_data(new_rollouts["data"])
            self.rollouts_info = new_rollouts["info"]

            if self.use_vtrace:
                compute_vtrace(self.ps.storage, self.ps.actor_critic, self.ps.algo)
            else:
                self.ps.storage.before_update(self.ps.actor_critic, self.ps.algo)

            # Prepare data batches
            self.batches = self.ps.storage.generate_batches(
                self.ps.algo.num_mini_batch, self.ps.algo.mini_batch_size,
                self.ps.algo.num_epochs, self.ps.actor_critic.is_recurrent)

        # Compute grads
        info = self.compute_gradients(self.batches.__next__())

        # Apply grads
        self.update_networks()

        # Add extra information to info dict
        info.update(self.rollouts_info)
        info.update({"scheme/metrics/gradient_update_delay": 0})
        info.update({"scheme/metrics/collection_gradient_delay": self.num_updates - self.rollouts_info["ac_version"]})

        # Update counter
        self.num_updates += 1
        self.rollouts_info["collected_samples"] = 0 # count only once

        return info

    def stop(self):
        """Stop remote workers"""
        for e in self.col_workers:
            e.terminate_worker.remote()

    def get_weights(self):
        """Returns current actor_critic.state_dict() weights"""
        return {k: v.cpu() for k, v in self.ps.actor_critic.state_dict().items()}

    def save_model(self, fname):
        """
        Save current version of actor_critic as a torch loadable checkpoint.

        Parameters
        ----------
        fname : str
            Filename given to the checkpoint.

        Returns
        -------
        save_name : str
            Path to saved file.
        """
        torch.save(self.ps.actor_critic.state_dict(), fname + ".tmp")
        os.rename(fname + '.tmp', fname)
        save_name = fname + ".{}".format(self.num_updates)
        copy2(fname, save_name)
        return save_name

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """
        self.ps.algo.update_algo_parameter(parameter_name, new_parameter_value)






import ray
from .c_worker import CWorker
from ..base.worker_set import WorkerSet as WS
from ..base.worker import default_remote_config

class CWorkerSet(WS):
    """
    Class to better handle the operations of ensembles of CWorkers.

    Parameters
    ----------
    create_algo_instance : func
        A function that creates an algorithm class.
    create_storage_instance : func
        A function that create a rollouts storage.
    create_train_envs_instance : func
        A function to create train environments.
    create_actor_critic_instance : func
        A function that creates a policy.
    create_test_envs_instance : func
        A function to create test environments.
    worker_remote_config : dict
        Ray resource specs for the remote workers.
    num_workers : int
        Number of remote workers in the worker set.

    Attributes
    ----------
    worker_class : python class
        Worker class to be instantiated to create Ray remote actors.
    remote_config : dict
        Ray resource specs for the remote workers.
    worker_params : dict
        Keyword arguments of the worker_class.
    num_workers : int
        Number of remote workers in the worker set.
    """
    def __init__(self,
                 create_algo_instance,
                 create_storage_instance,
                 create_train_envs_instance,
                 create_actor_critic_instance,
                 worker_remote_config=default_remote_config,
                 create_test_envs_instance=lambda x, y, c: None,
                 num_workers=1):

        self.worker_class = CWorker
        default_remote_config.update(worker_remote_config)
        self.remote_config = default_remote_config
        self.worker_params = {
            "create_algo_instance": create_algo_instance,
            "create_storage_instance": create_storage_instance,
            "create_test_envs_instance": create_test_envs_instance,
            "create_train_envs_instance": create_train_envs_instance,
            "create_actor_critic_instance": create_actor_critic_instance,
        }

        super(CWorkerSet, self).__init__(
            worker=self.worker_class,
            worker_params=self.worker_params,
            worker_remote_config=self.remote_config,
            num_workers=num_workers)



from baselines.common.atari_wrappers import make_atari, wrap_deepmind
from .wrappers import wrap_deepmind
from ..common import FrameStack

def make_atari_train_env(env_id, index_worker=0, index_env=0, seed=0, frame_stack=1):
    """
    Create train Atari environment.

    Parameters
    ----------
    env_id : str
        Environment name.
    index_worker : int
        Index of the worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    seed : int
        Environment random seed.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.

    Returns
    -------
    env : gym.Env
        Train environment.
    """
    env = make_atari(env_id)
    env.seed(seed + index_worker + index_env)
    env = wrap_deepmind(
        env, episode_life=True,
        clip_rewards=True,
        scale=False, frame_stack=1)

    if frame_stack > 1:
        env = FrameStack(env, frame_stack)

    return env

def make_atari_test_env(env_id, index_worker=0, index_env=0, seed=0, frame_stack=1):
    """
    Create test Atari environment.

    Parameters
    ----------
    env_id : str
        Environment name.
    index_worker : int
        Index of the worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    seed : int
        Environment random seed.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.

    Returns
    -------
    env : gym.Env
        Test environment.
    """
    env = make_atari(env_id)
    env.seed(seed + index_worker + index_env)
    env = wrap_deepmind(
        env, episode_life=False,
        clip_rewards=False,
        scale=False, frame_stack=1)

    if frame_stack > 1:
        env = FrameStack(env, frame_stack)

    return env


import os
import ray
import torch
from shutil import copy2
from collections import defaultdict

from .cgu_worker import CGUWorker
from .utils import broadcast_message
from ..base.worker_set import WorkerSet as WS
from ..base.worker import default_remote_config

class CGUWorkerSet(WS):
    """
    Class to better handle the operations of ensembles of workers.

    Parameters
    ----------
    create_algo_instance : func
        A function that creates an algorithm class.
    create_storage_instance : func
        A function that create a rollouts storage.
    create_train_envs_instance : func
        A function to create train environments.
    create_actor_critic_instance : func
        A function that creates a policy.
    create_test_envs_instance : func
        A function to create test environments.
    worker_remote_config : dict
        Ray resource specs for the remote workers.
    fraction_workers : float
        Minimum fraction of the remote workers to complete the data collection
        task to stop collection phase.
    fraction_samples : float
        Minimum fraction of the target steps allowed in a data collection.
    num_workers : int
        Number of remote workers in the worker set.

    Attributes
    ----------
    worker_class : python class
        Worker class to be instantiated to create Ray remote actors.
    remote_config : dict
        Ray resource specs for the remote workers.
    worker_params : dict
        Keyword arguments of the worker_class.
    num_workers : int
        Number of remote workers in the worker set.
    num_updates : int
        Times step() function has been called.
    fraction_workers : float
        Minimum fraction of the remote workers to complete the data collection
        task to stop collection phase.
    fraction_samples : float
        Minimum fraction of the target steps allowed in a data collection.
    """

    def __init__(self,
                 create_algo_instance,
                 create_storage_instance,
                 create_train_envs_instance,
                 create_actor_critic_instance,
                 create_test_envs_instance=lambda x, y, c: None,
                 worker_remote_config=default_remote_config,
                 fraction_workers=0.8,
                 fraction_samples=0.5,
                 num_workers=1):

        self.worker_class = CGUWorker
        default_remote_config.update(worker_remote_config)
        self.remote_config = default_remote_config
        self.worker_params = {
            "create_algo_instance": create_algo_instance,
            "create_storage_instance": create_storage_instance,
            "create_test_envs_instance": create_test_envs_instance,
            "create_train_envs_instance": create_train_envs_instance,
            "create_actor_critic_instance": create_actor_critic_instance,
        }

        super(CGUWorkerSet, self).__init__(
            worker=self.worker_class,
            worker_params=self.worker_params,
            worker_remote_config=self.remote_config,
            num_workers=num_workers)

        self.num_updates = 0
        self.num_workers = num_workers
        self.fraction_samples = fraction_samples if self.num_workers > 1 else 1.0
        self.fraction_workers = fraction_workers if self.num_workers > 1 else 1.0

        # Check remote workers exist
        if len(self.remote_workers()) == 0:
            raise ValueError("""at least 1 grad worker required""")

        # Setup the distributed processes for gradient averaging
        ip = ray.get(self.remote_workers()[0].get_node_ip.remote())
        port = ray.get(self.remote_workers()[0].find_free_port.remote())
        address = "tcp://{ip}:{port}".format(ip=ip, port=port)
        ray.get([worker.setup_torch_data_parallel.remote(
            address, i, len(self.remote_workers()), "nccl")
                 for i, worker in enumerate(self.remote_workers())])

    def step(self):
        """
        Takes a logical optimization step.

        Returns
        -------
        info : dict
            Summary dict of relevant information about the update process.
        """

        # Start data collection in all workers
        broadcast_message("sample", b"start-continue")

        pending_samples = [e.collect_data.remote(
            min_fraction=self.fraction_samples) for e in self.remote_workers()]

        # Keep checking how many workers have finished until percent% are ready
        samples_ready, samples_not_ready = ray.wait(
            pending_samples, num_returns=len(pending_samples), timeout=0.5)
        while len(samples_ready) < (self.num_workers * self.fraction_workers):
            samples_ready, samples_not_ready = ray.wait(
                pending_samples, num_returns=len(pending_samples), timeout=0.5)

        # Send stop message to the workers
        broadcast_message("sample", b"stop")

        # Compute model updates
        results = ray.get([e.step.remote() for e in self.remote_workers()])

        # Merge worker results
        step_metrics = defaultdict(float)
        for info in results:
            info["scheme/metrics/gradient_update_delay"] = self.num_updates - info.pop("ac_update_num")
            for k, v in info.items(): step_metrics[k] += v

        # Update info dict
        info = {k: v / self.num_workers for k, v in step_metrics.items()}

        # Update counters
        self.num_updates += 1

        return info

    def update_algo_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """
        for e in self.remote_workers():
            e.update_algo_parameter.remote(parameter_name, new_parameter_value)

    def save_model(self, fname):
        """
        Save current version of actor_critic as a torch loadable checkpoint.

        Parameters
        ----------
        fname : str
            Filename given to the checkpoint.

        Returns
        -------
        save_name : str
            Path to saved file.
        """
        model_dict = ray.get(self.remote_workers()[0].get_weights.remote())
        torch.save(model_dict, fname + ".tmp")
        os.rename(fname + '.tmp', fname)
        save_name = fname + ".{}".format(self.num_updates)
        copy2(fname, save_name)
        return save_name